var uuid = require('node-uuid');

function ProcessScanner(agent) {
  this.agent = agent;
  this.autoSnapshotDurationSeconds = 10;
  this.snapshotCount = 0;
  this.maxSnapshotsPerPeriod = 2;
  this.snapshotCountResetPeriodMS = 1 * 60 * 1000;
  this.snapshotsResetIntervalId = undefined;
  this.currentProcessSnapshot = undefined;
  this.lastProcessRequestId = undefined;
  this.btCallsStarted = null;
  this.btCallsCompleted = null;
}

exports.ProcessScanner = ProcessScanner;

ProcessScanner.prototype.init = function () {
  var self = this;

  var snapshotsPerPeriod = self.agent.opts.maxProcessSnapshotsPerPeriod;
  if (snapshotsPerPeriod) {
    self.maxSnapshotsPerPeriod = +snapshotsPerPeriod;
  }

  var resetPeriod = self.agent.opts.processSnapshotCountResetPeriodSeconds;
  if (resetPeriod) {
    self.snapshotCountResetPeriodMS = (+resetPeriod) * 1000;
  }

  self.agent.on('transactionStarted', function(transaction, req) {
    if (!self.capturingProcessSnapshot) return;
    self.btCallsStarted[transaction.id] = true;
    transaction.addProcessSnapshotGUID(self.currentProcessSnapshot.guid);
  });

  self.agent.on('transaction', function(transaction) {
    if (!self.capturingProcessSnapshot) return;

    transaction.addProcessSnapshotGUID(self.currentProcessSnapshot.guid);

    if (!self.btCallsStarted[transaction.id]) return;

    var id = transaction.registrationId || transaction.id;
    var btCall = self.btCallsCompleted[id] = self.btCallsCompleted[id] || { btID: id, count: 0, totalTimeTakenMS: 0 };
    btCall.totalTimeTakenMS += transaction.ms;
    btCall.count++;
  });
}

Object.defineProperty(ProcessScanner.prototype, 'capturingProcessSnapshot', {
    get: function(){
        return !!this.currentProcessSnapshot;
    }
});

function ensureInterval(self) {
  if (self.snapshotsResetIntervalId)
    return;
  self.snapshotsResetIntervalId = self.agent.timers.setInterval(function () {
    self.snapshotCount = 0;
  }, self.snapshotCountResetPeriodMS);
}

// translate an undefined callback to a function that
// logs errors and sends the process snapshot to the proxy
// on success.
function ensureCallback(self, callback)
{
  return callback || function (err, processSnapshot) {
    if (err) {
      self.agent.logger.error(err);
      return;
    }
    self.agent.proxyTransport.sendProcessSnapshot(processSnapshot);
  };
}

ProcessScanner.prototype.startAutoSnapshotIfPossible = function (callback) {
  var self = this;
  if (self.capturingProcessSnapshot) {
    self.agent.logger.log('Process snapshot already in progress, not starting another one!');
    return;
  }

  if (self.snapshotCount >= self.maxSnapshotsPerPeriod) {
    self.agent.logger.log(  'Already did '
                          + self.snapshotCount
                          + ' process snapshots in '
                          + self.snapshotCountResetPeriodMS
                          + 'ms.');
    return;
  }

  self.startSnapshot(self.autoSnapshotDurationSeconds, -1, ensureCallback(self, callback));
}

ProcessScanner.prototype.startManualSnapshot = function(processCallGraphReq, callback) {
  var self = this, requestID;

  requestID = +processCallGraphReq.snapshotRequestID;
  if(requestID <= self.lastSnapshotRequestID) {
    self.agent.logger.log('snapshotRequestID ' + processCallGraphReq.snapshotRequestID + ' was already processed, ignoring.')
    return;
  }

  self.lastSnapshotRequestID = requestID;

  if (self.capturingProcessSnapshot) {
    self.agent.logger.log('Process snapshot capture already in progress; snapshotRequestID ' + processCallGraphReq.snapshotRequestID + ' ignored.');
    return;
  }

  self.startSnapshot(processCallGraphReq.captureTime, requestID, ensureCallback(self, callback));
}

ProcessScanner.prototype.startSnapshot = function (durationSeconds,
                                                   requestID,
                                                   callback) {
  var self = this, requestID;

  if (self.capturingProcessSnapshot)
    return;

  self.btCallsStarted = {};
  self.btCallsCompleted = {};

  self.currentProcessSnapshot = {
      snapshotRequestID: requestID,
      guid : uuid.v4(),
      timestamp : undefined,
      processCallgraph : undefined,
      processID : process.pid,
      btCalls : undefined
  };
  self.snapshotCount++;
  ensureInterval(self);

  process.nextTick(function() {
    try {
      self.agent.cpuProfiler.startCpuProfiler(durationSeconds, function(err, processCallGraph) {
        try {
          if(err) {
            callback(err, null);
            return;
          }

          root = processCallGraph.callElements && processCallGraph.callElements[0];
          if (root && root.klass == "(global)" && root.method == "(root)") {
            processCallGraph.numOfRootElements = root.numChildren;
            processCallGraph.callElements.splice(0, 1);
          }

          var btCalls = Object.keys(self.btCallsCompleted).map(function(id) {
            return self.btCallsCompleted[id];
          });

          var processSnapshot = self.currentProcessSnapshot;
          processSnapshot.timestamp = Date.now();
          processSnapshot.processCallGraph = processCallGraph;
          processSnapshot.btCalls = btCalls;

          self.lastSnapshotRequestID = requestID;
          callback(null, processSnapshot);
        } finally {
          self.btCallsStarted = null;
          self.btCallsCompleted = null;
          self.currentProcessSnapshot = undefined;
        }
      });
      self.agent.profiler.processSnapshotStarted(self.currentProcessSnapshot);
    }
    catch(err) {
      self.btCallsStarted = null;
      self.btCallsCompleted = null;
      self.currentProcessSnapshot = undefined;
      callback(err, null);
      return;
    }
  });
}
